#include <iostream>
#include <cmath>
#include "main.h"

using namespace std;
using namespace irr;


// classes start

class CSampleSceneNode : public scene::ISceneNode
{

	core::aabbox3d<f32> Box;
	video::S3DVertex Vertices[128];
	video::SMaterial Material;


public:

	CSampleSceneNode(scene::ISceneNode* parent, scene::ISceneManager* mgr, s32 id)
		: scene::ISceneNode(parent, mgr, id)
	{
		Material.Wireframe = false;
		Material.Lighting = false;

		for(int i=0; i<128; i++){
            Vertices[i] = video::S3DVertex( 100*cos(i/128), 100*sin(i/128) , 100*cos(i/128), 1,1,0, video::SColor(floor(abs( 255*cos(3.14*i/128))),floor(abs( 255*sin(3.14*i/128))),floor(abs( 255*cos(3.14*i/128))),255), abs(cos(3.14*i/128)), abs(sin(3.14*i/128)));
		}

	/*
	The Irrlicht Engine needs to know the bounding box of your scene node.
	It will use it for doing automatic culling and other things. Hence we
	need to create a bounding box from the 4 vertices we use.
	If you do not want the engine to use the box for automatic culling,
	and/or don't want to create the box, you could also write
	AutomaticCullingEnabled = false;.
	*/
		Box.reset(Vertices[0].Pos);
		for (s32 i=1; i<128; ++i)
			Box.addInternalPoint(Vertices[i].Pos);
	}


	/*
	Before it is drawn, the OnRegisterSceneNode() method of every scene node in the scene
	is called by the scene manager. If the scene node wishes to draw itself,
	it may register itself in the scene manager to be drawn. This is necessary to
	tell the scene manager when it should call the ::render method. For example
	normal scene nodes render their content one after another, while
	stencil buffer shadows would like to be drawn after all other scene nodes. And
	camera or light scene nodes need to be rendered before all other scene
	nodes (if at all).
	So here we simply register the scene node to get render normally. If we would like
	to let it be rendered like cameras or light, we would have to call
	SceneManager->registerNodeForRendering(this, SNRT_LIGHT_AND_CAMERA);
	After this, we call the OnRegisterSceneNode-method of the base class ISceneNode,
	which simply lets also all the child scene nodes of this node register themselves.
	*/
	virtual void OnRegisterSceneNode()
	{
		if (IsVisible)
			SceneManager->registerNodeForRendering(this);

		ISceneNode::OnRegisterSceneNode();
	}

	/*
	In the render() method most of the interesting stuff happenes: The
	Scene node renders itself. We override this method and draw the
	tetraeder.
	*/
	virtual void render()
	{
		u16 indices[] = {	0,2,3, 2,1,3, 1,0,3, 2,0,1	};
		video::IVideoDriver* driver = SceneManager->getVideoDriver();

		driver->setMaterial(Material);
		driver->setTransform(video::ETS_WORLD, AbsoluteTransformation);
		driver->drawIndexedTriangleList(&Vertices[0], 4, &indices[0], 4);
	}

	/*
	At least, we create three small additional methods.
	GetBoundingBox() returns the bounding box of this scene node,
	GetMaterialCount() returns the amount of materials in this scene node
	(our tetraeder only has one material), and getMaterial() returns the
	material at an index. Because we have only one material here, we can
	return the only one meterial, assuming that no one ever calls getMaterial()
	with an index greater than 0.
	*/
	virtual const core::aabbox3d<f32>& getBoundingBox() const
	{
		return Box;
	}

	virtual u32 getMaterialCount() const
	{
		return 1;
	}

	virtual video::SMaterial& getMaterial(u32 i)
	{
		return Material;
	}
};


// classes end



int main(int argc, char* argv[])
{
    cout << "argc = " << argc << endl;
       for(int i = 0; i < argc; i++)
          cout << "argv[" << i << "] = " << argv[i] << endl;

    printf("Usage\n\nOA.exe [whoami] [server IP]\n\n[whoami] = 1/2/3\n[server IP] = IP del server o nomeComputer se messo dentro hosts.txt\n\n");

    ct = 0;

    if (enet_initialize () != 0)
    {
        fprintf (stderr, "An error occurred while initializing ENet.\n");
        return EXIT_FAILURE;
    }
    atexit (enet_deinitialize);

    whoami = atoi(argv[1]);

    if(whoami==1){
        server = new Server();
    } else {

        client = new Client();
        client->whoisclient = whoami;

        sprintf(client->indirizzo,"%s",argv[2]);
    }

    //FSOUND_Init(44100, 32, 0);
    //FSOUND_DSP_SetActive(FSOUND_DSP_GetFFTUnit(),true);

    video::E_DRIVER_TYPE driverType;

	printf("Please select the driver you want for this example:\n"\
		" (a) Direct3D 9.0c\n (b) Direct3D 8.1\n (c) OpenGL 1.5\n"\
		" (d) Software Renderer\n (e) Burning's Software Renderer\n"\
		" (f) NullDevice\n (otherKey) exit\n\n");

	char i;
	std::cin >> i;

	switch(i)
	{
		case 'a': driverType = video::EDT_DIRECT3D9;break;
		case 'b': driverType = video::EDT_DIRECT3D8;break;
		case 'c': driverType = video::EDT_OPENGL;   break;
		case 'd': driverType = video::EDT_SOFTWARE; break;
		case 'e': driverType = video::EDT_BURNINGSVIDEO;break;
		case 'f': driverType = video::EDT_NULL;     break;
		default: return 0;
	}

	printf("select mode:\n"\
        "(a) OneAvatar\n"\
        "(b) Player 1\n"\
        "(c) Player 2.\n\n");

    std::cin >> i;

	switch(i)
	{
		case 'a': whoami=1;break;
		case 'b': whoami=2;break;
		case 'c': whoami=3;break;
		default: whoami=1;
	}

	device = createDevice(driverType,core::dimension2d<s32>(1024, 768),16, true, false, true, &receiver);

	if (device == 0)
		return 1; // could not create selected driver.

	device->setWindowCaption(L"OneAvatar");

	video::IVideoDriver* driver = device->getVideoDriver();
	scene::ISceneManager* smgr = device->getSceneManager();


    setup(smgr,device,driver);

    device->getCursorControl()->setVisible(false);

    loop(smgr,device,driver);



    return 0;

}


void setup(scene::ISceneManager* smgr,IrrlichtDevice *device,video::IVideoDriver* driver){

    smgr->addCameraSceneNodeFPS();

    CSampleSceneNode *myNode =
        new CSampleSceneNode(smgr->getRootSceneNode(), smgr, 666);
    scene::ISceneNodeAnimator* anim =
        smgr->createRotationAnimator(core::vector3df(0.8f, 0, 0.8f));

    device->getFileSystem()->addZipFileArchive("mappa.pk3");

    scene::IAnimatedMesh* q3levelmesh = smgr->getMesh("20kdm2.bsp");
    q3node = 0;

    if (q3levelmesh)
            q3node = smgr->addOctTreeSceneNode(q3levelmesh->getMesh(0));


    selector = 0;

    if (q3node)
    {
            q3node->setPosition(core::vector3df(0,0,0));

            selector = smgr->createOctTreeTriangleSelector(q3levelmesh->getMesh(0), q3node, 128);
            q3node->setTriangleSelector(selector);
    }

    camera =smgr->addCameraSceneNodeFPS(0, 100.0f, 300.0f, -1, 0, 0, false);
    camera->setPosition(core::vector3df(976,164,1346 ));
    camera->setTarget(core::vector3df(1404,164,1346));

    if (selector)
    {
            scene::ISceneNodeAnimator* anim = smgr->createCollisionResponseAnimator(
                    selector, camera, core::vector3df(30,50,30),
                    core::vector3df(0,-2,0),
                    core::vector3df(0,0,0));
            camera->addAnimator(anim);
            anim->drop();
    }


        bill = smgr->addBillboardSceneNode();
        bill->setMaterialType(video::EMT_TRANSPARENT_ADD_COLOR );
        bill->setMaterialTexture(0, driver->getTexture("particle.bmp"));
        bill->setMaterialFlag(video::EMF_LIGHTING, false);
        bill->setMaterialFlag(video::EMF_ZBUFFER, false);
        bill->setSize(core::dimension2d<f32>(20.0f, 20.0f));

        material.setTexture(0, driver->getTexture("supermale/OA.jpg"));
        material.Lighting = false;

        materialWhite.setTexture(0, driver->getTexture("supermale/OAWhite.jpg"));
        materialWhite.Lighting = false;

        node1 = 0;
        node2 = 0;
        node3 = 0;
        scene::IAnimatedMesh* faerie = smgr->getMesh("supermale/tris.md2");

        if (faerie)
        {

                if( whoami!=1 ){
                    // signor 1 start
                    node1 = smgr->addAnimatedMeshSceneNode(faerie);
                    node1->setPosition(core::vector3df(1404,94,1346));
                    node1->setMD2Animation(scene::EMAT_RUN);
                    node1->getMaterial(0) = material;


                    /*
                    scene::ISceneNodeAnimator* anim = smgr->createCollisionResponseAnimator(
                        selector, node1, core::vector3df(30,50,30),
                        core::vector3df(0,-2,0),
                        core::vector3df(0,50,0));

                    node1->addAnimator(anim);
                    anim->drop();
                    */
                    // signor 1 end
                }

                anim = 0;

                if( whoami!=2 ){
                    // signor 2 start
                    node2 = smgr->addAnimatedMeshSceneNode(faerie);
                    node2->setPosition(core::vector3df(1404,94,1316));
                    node2->setMD2Animation(scene::EMAT_STAND);
                    node2->getMaterial(0) = materialWhite;

                    /*

                    anim = smgr->createCollisionResponseAnimator(
                        selector, node2, core::vector3df(30,50,30),
                        core::vector3df(0,-2,0),
                        core::vector3df(0,50,0));

                    node2->addAnimator(anim);
                    anim->drop();
                    */
                    // signor 2 end
                }

                anim = 0;

                if( whoami!=3){
                    // signor 3 start
                    node3 = smgr->addAnimatedMeshSceneNode(faerie);
                    node3->setPosition(core::vector3df(1404,94,1376));
                    node3->setMD2Animation(scene::EMAT_STAND);
                    node3->getMaterial(0) = materialWhite;

                    /*
                    anim = smgr->createCollisionResponseAnimator(
                        selector, node3, core::vector3df(30,50,30),
                        core::vector3df(0,-2,0),
                        core::vector3df(0,50,0));

                    node3->addAnimator(anim);
                    anim->drop();
                    */
                    // signor 3 end
                }

                anim = 0;

        }

        if(selector)
            selector->drop();

        material.setTexture(0, 0);
        material.Lighting = false;

        // Add a light

        smgr->addLightSceneNode(0, core::vector3df(1404,134,1346),
                video::SColorf(1.0f,1.0f,1.0f,1.0f),
                600.0f);


    //bkgSoundLounge = FSOUND_Stream_Open("mex_01.mp3",FSOUND_LOOP_NORMAL,0,0);
    //FSOUND_Stream_Play(FSOUND_FREE , bkgSoundLounge);


}


void loop(scene::ISceneManager* smgr,IrrlichtDevice *device, video::IVideoDriver* driver){

    scene::ISceneNode* selectedSceneNode = 0;
        scene::ISceneNode* lastSelectedSceneNode = 0;


        int lastFPS = -1;


    while(device->run())
	if (device->isWindowActive())
	{



        ct = (ct + 1)%5;

            if(whoami!=1 && !client->connected){
                client->connect();
            }

            if(whoami==1 && server){
                server->p1x = camera->getPosition().X;
                server->p1y = camera->getPosition().Y-25;
                server->p1z = camera->getPosition().Z;
                server->r1x = 0;//camera->getRotation().X;
                server->r1y = camera->getRotation().Y-90;
                server->r1z = 0;//camera->getRotation().Z;
                server->procEv();
            } else {
                if(whoami==2){
                    client->p2x = camera->getPosition().X;
                    client->p2y = camera->getPosition().Y-25;
                    client->p2z = camera->getPosition().Z;
                    client->r2x = 0;//camera->getRotation().X;
                    client->r2y = camera->getRotation().Y-90;
                    client->r2z = 0;//camera->getRotation().Z;
                } else if(whoami==3) {
                    client->p3x = camera->getPosition().X;
                    client->p3y = camera->getPosition().Y-25;
                    client->p3z = camera->getPosition().Z;
                    client->r3x = 0;//camera->getRotation().X;
                    client->r3y = camera->getRotation().Y-90;
                    client->r3z = 0;//camera->getRotation().Z;
                }

             }

             if(whoami==1 && server){
                if(node2){
                    node2->setPosition( core::vector3df( server->p2x,server->p2y,server->p2z ) );
                    node2->setRotation( core::vector3df( server->r2x,server->r2y,server->r2z ) );
                }
                if(node3){
                    node3->setPosition( core::vector3df( server->p3x,server->p3y,server->p3z ) );
                    node3->setRotation( core::vector3df( server->r3x,server->r3y,server->r3z ) );
                }
             } else if(whoami==2 && client){
                 if(node1){
                    node1->setPosition( core::vector3df( client->p1x,client->p1y,client->p1z ) );
                    node1->setRotation( core::vector3df( client->r1x,client->r1y,client->r1z ) );
                }
                if(node3){
                    node3->setPosition( core::vector3df( client->p3x,client->p3y,client->p3z ) );
                    node3->setRotation( core::vector3df( client->r3x,client->r3y-45,client->r3z ) );
                }
             } else if(whoami==3 && client){
                 if(node1){
                    node1->setPosition( core::vector3df( client->p1x,client->p1y,client->p1z ) );
                    node1->setRotation( core::vector3df( client->r1x,client->r1y-45,client->r1z ) );
                }
                if(node2){
                    node2->setPosition( core::vector3df( client->p2x,client->p2y,client->p2z ) );
                    node2->setRotation( core::vector3df( client->r2x,client->r2y-45,client->r2z ) );
                }

             }


                if(ct==0 && whoami!=1 ){
                    client->updCoords();
                    client->procEv();
                }

	   //float* spectrum =   FSOUND_DSP_GetSpectrum();


		driver->beginScene(true, true, video::SColor(0,200,200,200));

		smgr->drawAll();

                core::line3d<f32> line;
                line.start = camera->getPosition();
                line.end = line.start + (camera->getTarget() - line.start).normalize() * 1000.0f;

                core::vector3df intersection;
                core::triangle3df tri;

                if (smgr->getSceneCollisionManager()->getCollisionPoint(
                        line, selector, intersection, tri))
                {
                        bill->setPosition(intersection);


                        //driver->setTransform(video::ETS_WORLD, core::matrix4());
                        //driver->setMaterial(material);
                     //   driver->draw3DTriangle(tri, video::SColor(0,255,0,0));
                }

                selectedSceneNode = smgr->getSceneCollisionManager()->getSceneNodeFromCameraBB(camera);

                if (lastSelectedSceneNode && lastSelectedSceneNode==node1 && whoami!=1 ){
                    node1->setMD2Animation(scene::EMAT_STAND);
                    if( whoami!=1 && client){
                        client->hasHit = 0;
                    }
                }
                        //lastSelectedSceneNode->setMaterialFlag(video::EMF_LIGHTING, false);

                if (selectedSceneNode == q3node || selectedSceneNode == bill )
                        selectedSceneNode = 0;

                if (selectedSceneNode && selectedSceneNode==node1 && whoami!=1 ){
                        srand((unsigned)time(0));
                        int r = (rand()%3);
                        if(r==0){
                            node1->setMD2Animation(scene::EMAT_PAIN_A);
                        } else if(r==1){
                            node1->setMD2Animation(scene::EMAT_PAIN_B);
                        } else if(r==2){
                            node1->setMD2Animation(scene::EMAT_PAIN_C);
                        }
                        if( whoami!=1 && client){
                            client->hasHit = 1;
                        }
                        if(client){
                            client->sendShock();
                        }
                        //selectedSceneNode->setMaterialFlag(video::EMF_LIGHTING, true);
                }

                lastSelectedSceneNode = selectedSceneNode;

		driver->endScene();

        int fps = driver->getFPS();

                if (lastFPS != fps)
                {
                        core::stringw str = L"Collision detection example - Irrlicht Engine [";
                        str += driver->getName();
                        str += "] FPS:";
                        str += fps;
                        str += "x=";
                        str += camera->getPosition().X;
                        str += "y=";
                        str += camera->getPosition().Y;
                        str += "z=";
                        str += camera->getPosition().Z;

                        device->setWindowCaption(str.c_str());
                        lastFPS = fps;
                }

	}

	device->drop();

}


